from __future__ import annotations
import json

from websockets.sync.client import connect, ClientConnection

from .logger import logger
from .model.ogmios_model import Jsonrpc
from .errors import InvalidResponseError

from .chainsync import FindIntersection, NextBlock
from .statequery import (
    AcquireLedgerState,
    QueryBlockHeight,
    QueryNetworkTip,
    QueryStartTime,
    QueryLedgerTip,
    QueryGenesisConfiguration,
    QueryEpoch,
    QueryEraStart,
    QueryEraSummaries,
    QueryLiveStakeDistribution,
    QueryStakePools,
    QueryUtxo,
    QueryProjectedRewards,
    QueryRewardsProvenance,
    QueryRewardAccountSummaries,
    QueryProtocolParameters,
    QueryProposedProtocolParameters,
    QueryConstitution,
    QueryConstitutionalCommittee,
    QueryTreasuryAndReserves,
)
from .mempool import AcquireMempool, ReleaseMempool, SizeOfMempool, HasTransaction, NextTransaction
from .txsubmit import SubmitTransaction, EvaluateTransaction


class Client:
    """
    Ogmios connection client

    A subset of Ogmios functions require the use of WebSockets. Therefore a
    WebSocket connection is preferred over HTTP. If http_only is set to True,
    functions that require WebSockets will not be available.

    If secure is set to False, ws / http will be used rather than wss / https

    :param host: The host of the Ogmios server
    :type host: str
    :param port: The port of the Ogmios server
    :type port: int
    :param secure: Use secure connection
    :type secure: bool
    :param http_only: Use HTTP connection
    :type http_only: bool
    :param compact: Use compact connection
    :type compact: bool
    :param rpc_version: The JSON-RPC version to use
    :type rpc_version: Jsonrpc
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 1337,
        path: str = "",
        secure: bool = False,
        http_only: bool = False,
        rpc_version: Jsonrpc = Jsonrpc.field_2_0,
        additional_headers: dict = {},
    ) -> None:
        if http_only:
            protocol: str = "https" if secure else "http"
            # TODO: Implement HTTP connections
            logger.error("HTTP connections not implemented")
            raise NotImplementedError
        else:
            protocol: str = "wss" if secure else "ws"

        self.rpc_version = rpc_version
        connect_str: str = f"{protocol}://{host}:{port}/{path}"
        self.connection: ClientConnection = connect(
            connect_str, additional_headers=additional_headers
        )

        # Ogmios chainsync methods
        self.find_intersection = FindIntersection(self)
        self.next_block = NextBlock(self)

        # Ogmios statequery methods
        self.acquire_ledger_state = AcquireLedgerState(self)
        self.query_block_height = QueryBlockHeight(self)
        self.query_network_tip = QueryNetworkTip(self)
        self.query_ledger_tip = QueryLedgerTip(self)
        self.query_start_time = QueryStartTime(self)
        self.query_genesis_configuration = QueryGenesisConfiguration(self)
        self.query_epoch = QueryEpoch(self)
        self.query_era_start = QueryEraStart(self)
        self.query_era_summaries = QueryEraSummaries(self)
        self.query_live_stake_distribution = QueryLiveStakeDistribution(self)
        self.query_stake_pools = QueryStakePools(self)
        self.query_utxo = QueryUtxo(self)
        self.query_projected_rewards = QueryProjectedRewards(self)
        self.query_rewards_provenance = QueryRewardsProvenance(self)
        self.query_reward_account_summaries = QueryRewardAccountSummaries(self)
        self.query_protocol_parameters = QueryProtocolParameters(self)
        self.query_proposed_protocol_parameters = QueryProposedProtocolParameters(self)
        self.query_constitution = QueryConstitution(self)
        self.query_constitutional_committee = QueryConstitutionalCommittee(self)
        self.query_treasury_and_reserves = QueryTreasuryAndReserves(self)

        # Ogmios mempool methods
        self.acquire_mempool = AcquireMempool(self)
        self.release_mempool = ReleaseMempool(self)
        self.size_of_mempool = SizeOfMempool(self)
        self.has_transaction = HasTransaction(self)
        self.next_transaction = NextTransaction(self)

        # Ogmios txsubmit methods
        self.submit_transaction = SubmitTransaction(self)
        self.evaluate_transaction = EvaluateTransaction(self)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        """Close client connection when finished"""
        self.connection.close()

    def send(self, request: str) -> None:
        """Send a request to the Ogmios server

        :param request: The request to send
        :type request: str
        """
        self.connection.send(request)

    def receive(self) -> dict:
        """Receive a response from the Ogmios server

        :return: Request response
        """
        resp = json.loads(self.connection.recv())
        if resp.get("version"):
            raise InvalidResponseError(
                f"Invalid Ogmios version. ogmios-python only supports Ogmios server version v6.0.0 and above."
            )
        return resp



from __future__ import annotations

from typing import Any, List, Optional, Tuple, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ogmios.client import Client

from ogmios.errors import InvalidMethodError, InvalidResponseError, ResponseError
from ogmios.logger import logger
from ogmios.datatypes import TxOutputReference, Address, Utxo, Script
import ogmios.model.ogmios_model as om
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


class QueryUtxo:
    """Ogmios method to query the current UTxO set, restricted to some output references or
    addresses.

    NOTE: This class is not intended to be used directly. Instead, use the Client.query_utxo
    method.

    :param client: The client to use for the request.
    :type client: Client
    """

    def __init__(self, client: Client):
        self.client = client
        self.method = mm.Method.queryLedgerState_utxo.value

    def execute(
        self, output_ref: Union[List[TxOutputReference], List[Address]], id: Optional[Any] = None
    ) -> Tuple[list[Utxo], Optional[Any]]:
        """Send and receive the request.

        :param output_ref: The output reference to query. Can be a list of TxOutputReference or a list
            of Address.
        :type output_ref: list[TxOutputReference] | list[Address]
        :param id: The ID of the request.
        :type id: Any
        :return: List of Utxo's and ID of the response.
        :rtype: (list[Utxo], Optional[Any])
        """
        self.send(output_ref, id)
        return self.receive()

    def send(
        self, output_ref: Union[List[TxOutputReference], List[Address]], id: Optional[Any] = None
    ) -> None:
        """Send the request.

        :param output_ref: The output reference to query. Can be a list of TxOutputReference or a list
            of Address.
        :param id: The ID of the request.
        :type id: Any
        """
        if isinstance(output_ref, list) and all(isinstance(ref, Address) for ref in output_ref):
            addresses = [addr.address for addr in output_ref]
            params = om.Params8(addresses=addresses)
        elif isinstance(output_ref, list) and all(
            isinstance(ref, TxOutputReference) for ref in output_ref
        ):
            outputReferences = [ref._schematype for ref in output_ref]
            params = om.Params7(outputReferences=outputReferences)
        else:
            raise TypeError(
                f"Invalid type for output_ref: {type(output_ref)}. Must be Utxo, list[Address], or list[TxOutputReference]"
            )

        pld = om.QueryLedgerStateUtxo(
            jsonrpc=self.client.rpc_version,
            method=self.method,
            params=params,
            id=id,
        )
        self.client.send(pld.json())

    def receive(self) -> Tuple[List[Utxo], Optional[Any]]:
        """Receive the response.

        :return: List of Utxo's and ID of the response.
        :rtype: (list[Utxo], Optional[Any])
        """
        response = self.client.receive()
        return self._parse_QueryUtxo_response(response)

    @staticmethod
    def _parse_QueryUtxo_response(
        response: dict,
    ) -> Tuple[List[Utxo], Optional[Any]]:
        if response.get("method") != mm.Method.queryLedgerState_utxo.value:
            raise InvalidMethodError(f"Incorrect method for query_utxo response: {response}")

        if response.get("error"):
            raise ResponseError(f"Ogmios responded with error: {response}")

        id: Optional[Any] = response.get("id")

        if response.get("result") == []:
            logger.info("No UTxOs found")
            return [], id

        if result := response.get("result"):
            utxos = []
            for utxo in result:
                utxos.append(
                    Utxo(
                        tx_id=utxo.get("transaction").get("id"),
                        index=utxo.get("index"),
                        address=utxo.get("address"),
                        value=utxo.get("value"),
                        datum_hash=utxo.get("datumHash"),
                        datum=utxo.get("datum"),
                        script=utxo.get("script"),
                    )
                )
            logger.info(
                f"""Parsed utxo response:
        UTxOs = {utxos}
        ID = {id}"""
            )
            return utxos, id

        raise InvalidResponseError(f"Failed to parse query_utxo response: {response}")



class InvalidMethodError(Exception):
    """Raised when Ogmios response contains an unexpected method name"""

    pass


class InvalidResponseError(Exception):
    """Raised when Ogmios response contains unexpected content"""

    pass


class InvalidOgmiosParameter(Exception):
    """Raised when missing or invalid parameters are passed to an Ogmios object"""

    pass


class ResponseError(Exception):
    """Raised when an Ogmios request contains an error message"""

    pass



from datetime import datetime
import pytest

from ogmios.datatypes import TxOutputReference, Address, Utxo
from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryUtxo_address_empty(client):
    addr = Address(
        "addr1q9d34spgg2kdy47n82e7x9pdd6vql6d2engxmpj20jmhuc2047yqd4xnh7u6u5jp4t0q3fkxzckph4tgnzvamlu7k5psuahzcp"
    )
    id_str = "My ID string"
    utxos, id = client.query_utxo.execute([addr], id_str)

    assert utxos == []
    assert id == id_str


def test_QueryUtxo_address(client):
    # This test relies on this address having UTxO's on the testnet
    addr = Address("addr_test1vz09v9yfxguvlp0zsnrpa3tdtm7el8xufp3m5lsm7qxzclgmzkket")
    id_str = "My ID string"
    utxos, id = client.query_utxo.execute([addr], id_str)

    assert isinstance(utxos, list)
    assert len(utxos) > 0
    assert isinstance(utxos[0], Utxo)
    assert id == id_str


@pytest.mark.skip
def test_QueryUtxo_output_ref(client):
    output_ref = TxOutputReference(
        tx_id="eac76c86b2c8552204b89fdc6f1506f59a6b464ecebd5468aa008ce0fa95c5e8", index=0
    )
    id_str = "My ID string"
    utxos, id = client.query_utxo.execute([output_ref], id_str)
    utxo = utxos[0]

    assert isinstance(utxos, list)
    assert utxo.tx_id == output_ref.tx_id
    assert utxo.index == output_ref.index
    assert (
        utxo.address
        == 'addr_test1qznwxwlxrq9p58c4mfmxdh45zpmu5wfhflwhp69cq7v667h7cvpcq5rkwhjpjmandtq39es7yxvluyts9kfgjcajg24sx9yfqe'
    )
    assert utxo.value.get("ada").get("lovelace") == 11000000
    assert len(utxos) > 0
    assert isinstance(utxos[0], Utxo)
    assert id == id_str



# SPDX-FileCopyrightText: 2023-present Willie Marchetto <willie@viperscience.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later
from .logger import logger
from .client import Client
from .datatypes import (
    Origin,
    Point,
    Tip,
    Block,
    GenesisConfiguration,
    Era,
    Direction,
    ProtocolParameters,
    AlonzoUpdatableParameters,
    ConwayUpdatableParameters,
    BootstrapProtocolParameters,
    GovernanceProtocolParameters,
    Ada,
    Script,
    Utxo,
    TxOutputReference,
    Address,
)
from .chainsync import FindIntersection, NextBlock
from .statequery import (
    AcquireLedgerState,
    QueryBlockHeight,
    QueryNetworkTip,
    QueryStartTime,
    QueryLedgerTip,
    QueryEpoch,
    QueryEraStart,
    QueryEraSummaries,
    QueryLiveStakeDistribution,
    QueryStakePools,
    QueryUtxo,
    QueryProjectedRewards,
    QueryRewardsProvenance,
    QueryRewardAccountSummaries,
    QueryProtocolParameters,
    QueryProposedProtocolParameters,
    QueryConstitution,
    QueryConstitutionalCommittee,
    QueryTreasuryAndReserves,
)
from .mempool import AcquireMempool, ReleaseMempool, SizeOfMempool, HasTransaction, NextTransaction
from .txsubmit import SubmitTransaction, EvaluateTransaction
from .utils import (
    GenesisParameters,
    get_current_era,
    wait_for_empty_mempool,
    get_mempool_transactions,
)
from .errors import InvalidOgmiosParameter, InvalidMethodError, InvalidResponseError, ResponseError

__all__ = [
    "logger",
    "Client",
    "Origin",
    "Point",
    "Tip",
    "Block",
    "GenesisConfiguration",
    "Era",
    "ProtocolParameters",
    "AlonzoUpdatableParameters",
    "ConwayUpdatableParameters",
    "BootstrapProtocolParameters",
    "GovernanceProtocolParameters",
    "Ada",
    "Script",
    "Utxo",
    "TxOutputReference",
    "Address",
    "Direction",
    "FindIntersection",
    "NextBlock",
    "InvalidOgmiosParameter",
    "InvalidMethodError",
    "InvalidResponseError",
    "ResponseError",
    "AcquireLedgerState",
    "QueryBlockHeight",
    "QueryNetworkTip",
    "QueryStartTime",
    "QueryLedgerTip",
    "QueryEpoch",
    "QueryEraStart",
    "QueryEraSummaries",
    "QueryLiveStakeDistribution",
    "QueryStakePools",
    "QueryUtxo",
    "QueryProjectedRewards",
    "QueryRewardsProvenance",
    "QueryRewardAccountSummaries",
    "QueryProtocolParameters",
    "QueryProposedProtocolParameters",
    "AcquireMempool",
    "ReleaseMempool",
    "SizeOfMempool",
    "HasTransaction",
    "NextTransaction",
    "SubmitTransaction",
    "EvaluateTransaction",
    "GenesisParameters",
    "get_current_era",
    "wait_for_empty_mempool",
    "get_mempool_transactions",
]



import pytest

from ogmios.errors import ResponseError
from tests.ogmios.test_fixtures import client


# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_ReleaseMempool(client):
    id_str = "My ID string"

    with pytest.raises(ResponseError):
        # Try releasing a mempool snapshot before acquiring one
        id = client.release_mempool.execute(id_str)

    # Acquire a mempool snapshot
    client.acquire_mempool.execute()
    id = client.release_mempool.execute(id_str)
    assert id == id_str



from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryStartTime(client):
    id_str = "My ID string"
    stake_dist, id = client.query_live_stake_distribution.execute(id_str)

    assert isinstance(stake_dist, dict)
    assert id == id_str

    # Verify dict entries have the proper keys
    first_pool = list(stake_dist.keys())[0]
    assert stake_dist.get(first_pool).keys() == {"stake", "vrf"}



from __future__ import annotations

from typing import Any, Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ogmios.client import Client

from ogmios.errors import InvalidMethodError, InvalidResponseError, ResponseError
from ogmios.logger import logger
from ogmios.datatypes import EraSummary
import ogmios.model.ogmios_model as om
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


class QueryEraSummaries:
    """Ogmios method to query a summary of the slotting parameters and boundaries for each known
    era. Useful for doing slot-arithmetic and time conversions.

    NOTE: This class is not intended to be used directly. Instead, use the
    Client.query_era_summaries method.

    :param client: The client to use for the request.
    :type client: Client
    """

    def __init__(self, client: Client):
        self.client = client
        self.method = mm.Method.queryLedgerState_eraSummaries.value

    def execute(self, id: Optional[Any] = None) -> (list[EraSummary], Optional[Any]):
        """Send and receive the request.

        :param id: The ID of the request.
        :type id: Any
        :return: List of era summaries and ID of the response.
        :rtype: (list[EraSummary], Optional[Any])
        """
        self.send(id)
        return self.receive()

    def send(self, id: Optional[Any] = None) -> None:
        """Send the request.

        :param id: The ID of the request.
        :type id: Any
        """
        pld = om.QueryLedgerStateEraSummaries(
            jsonrpc=self.client.rpc_version,
            method=self.method,
            id=id,
        )
        self.client.send(pld.json())

    def receive(self) -> (int, int, int, Optional[Any]):
        """Receive a previously requested response.

        :return: List of era summaries and ID of the response.
        :rtype: (list[eraSummary], Optional[Any])
        """
        response = self.client.receive()
        return self._parse_QueryEraSummaries_response(response)

    @staticmethod
    def _parse_QueryEraSummaries_response(response: dict) -> (list[EraSummary], Optional[Any]):
        if response.get("method") != mm.Method.queryLedgerState_eraSummaries.value:
            raise InvalidMethodError(
                f"Incorrect method for query_era_summaries response: {response}"
            )

        if response.get("error"):
            raise ResponseError(f"Ogmios responded with error: {response}")

        if result := response.get("result"):
            summaries = []
            for era_summary in result:
                if not isinstance(era_summary, dict):
                    raise InvalidResponseError(
                        f"Failed to parse query_era_summaries response: {response}"
                    )
                if (
                    (start := era_summary.get("start")) is not None
                    and (end := era_summary.get("end")) is not None
                    and (parameters := era_summary.get("parameters")) is not None
                ):
                    summaries.append(
                        EraSummary(
                            start_time=start.get("time").get("seconds"),
                            start_slot=start.get("slot"),
                            start_epoch=start.get("epoch"),
                            end_time=end.get("time").get("seconds"),
                            end_slot=end.get("slot"),
                            end_epoch=end.get("epoch"),
                            epoch_length=parameters.get("epochLength"),
                            slot_length=parameters.get("slotLength").get("milliseconds"),
                            safe_zone=parameters.get("safeZone"),
                        )
                    )
            id: Optional[Any] = response.get("id")
            logger.info(
                f"""Parsed query_era_start response:
EraSummaries = {summaries}
ID = {id}"""
            )
            return summaries, id
        raise InvalidResponseError(f"Failed to parse query_era_summaries response: {response}")



import pytest

from ogmios.errors import ResponseError
from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_HasTransaction(client):
    id_str = "My ID string"
    tx_id = "90181c517248757d3a5beadc9c3fe64bf821d3e889a963fc717003ec248757d3"
    with pytest.raises(ResponseError):
        # Try getting mempool size before acquiring a snapshot
        client.has_transaction.execute(tx_id)

    # Acquire a mempool snapshot
    client.acquire_mempool.execute()
    has_tx, id = client.has_transaction.execute(tx_id, id_str)
    assert isinstance(has_tx, bool)
    assert not has_tx
    assert id == id_str



import pytest
from typing import Union

from ogmios import ProtocolParameters
from tests.ogmios.test_fixtures import client


@pytest.mark.skip
def test_QueryProposedProtocolParameters(client):
    id_str = "My ID string"
    result, id = client.query_proposed_protocol_parameters.execute(id_str)

    assert isinstance(result, Union[ProtocolParameters, None])
    assert id == id_str



import pytest
from datetime import datetime
from pydantic.v1.error_wrappers import ValidationError

from ogmios.datatypes import (
    Origin,
    Point,
    Tip,
    Block,
    GenesisConfiguration,
    ProtocolParameters,
    BootstrapProtocolParameters,
    Script,
)
from ogmios.model.model_map import Types
import ogmios.model.ogmios_model as om

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_Origin():
    # Valid origin
    origin = Origin()
    assert origin._schematype == om.Origin.origin


def test_Point():
    # Valid Point
    Point(slot=10, id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af")

    # Invalid slot
    with pytest.raises(ValidationError):
        Point(slot=-10, id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af")

    # Invalid id
    with pytest.raises(ValidationError):
        Point(slot=10, id="bad_id")


def test_Tip():
    # Valid Tip
    Tip(slot=10, height=1000, id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af")

    # Invalid slot
    with pytest.raises(ValidationError):
        Tip(
            slot=-10,
            height=1000,
            id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        )

    # Invalid height
    with pytest.raises(ValidationError):
        Tip(
            slot=10,
            height=-1000,
            id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        )

    # Invalid id
    with pytest.raises(ValidationError):
        Tip(slot=10, height=1000, id="bad_id")


def test_Tip_to_point():
    tip = Tip(
        slot=10, height=1000, id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
    )
    point = Point(slot=10, id="45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af")
    assert tip.to_point() == point


def test_Block_EBB():
    # Valid EBB block
    kwargs = {
        "era": "byron",
        "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        "height": 100,
    }
    Block(Types.ebb.value, **kwargs)

    # Wrong era
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "shelley",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "height": 100,
        }
        Block(Types.ebb.value, **kwargs)

    # Invalid height
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "byron",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "height": -100,
        }
        Block(Types.ebb.value, **kwargs)

    # Missing required parameter
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "byron",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "height": 100,
        }
        Block(Types.ebb.value, **kwargs)


def test_Block_BFT():
    # Valid block
    kwargs = {
        "era": "byron",
        "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        "height": 100,
        "slot": 1000,
        "size": {"bytes": 1},
        "transactions": [],  # TODO: Add this type
        "protocol": {
            "id": 123456,
            "version": {"major": 0, "minor": 1, "patch": 2},
            "software": {"appName": "test", "number": 1234},
        },
        "issuer": {
            "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
        },
        "delegate": {
            "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
        },
    }
    Block(Types.bft.value, **kwargs)

    # Wrong era
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "shelley",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "height": 100,
            "slot": 1000,
            "size": {"bytes": 1},
            "transactions": [],  # TODO: Add this type
            "protocol": {
                "id": 123456,
                "version": {"major": 0, "minor": 1, "patch": 2},
                "software": {"appName": "test", "number": 1234},
            },
            "issuer": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
            },
            "delegate": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
            },
        }
        Block(Types.bft.value, **kwargs)

    # Invalid height
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "byron",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "height": -100,
            "slot": 1000,
            "size": {"bytes": 1},
            "transactions": [],  # TODO: Add this type
            "protocol": {
                "id": 123456,
                "version": {"major": 0, "minor": 1, "patch": 2},
                "software": {"appName": "test", "number": 1234},
            },
            "issuer": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
            },
            "delegate": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
            },
        }
        Block(Types.bft.value, **kwargs)

    # Missing required parameter
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "byron",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "slot": 1000,
            "size": {"bytes": 1},
            "transactions": [],  # TODO: Add this type
            "protocol": {
                "id": 123456,
                "version": {"major": 0, "minor": 1, "patch": 2},
                "software": {"appName": "test", "number": 1234},
            },
            "issuer": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
            },
            "delegate": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af"
            },
        }
        Block(Types.bft.value, **kwargs)


def test_Block_Praos():
    # Valid block
    kwargs = {
        "era": "alonzo",
        "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        "nonce": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
        "height": 100,
        "slot": 1000,
        "size": {"bytes": 1},
        "transactions": [],  # TODO: Add this type
        "protocol": {
            "version": {"major": 0, "minor": 1, "patch": 2},
        },
        "issuer": {
            "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "vrfVerificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "operationalCertificate": {
                "count": 123,
                "kes": {
                    "period": 12,
                    "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                },
            },
            "leaderValue": {},
        },
    }
    Block(Types.praos.value, **kwargs)

    # Wrong era
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "byron",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "nonce": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "height": 100,
            "slot": 1000,
            "size": {"bytes": 1},
            "transactions": [],  # TODO: Add this type
            "protocol": {
                "version": {"major": 0, "minor": 1, "patch": 2},
            },
            "issuer": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                "vrfVerificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                "operationalCertificate": {
                    "count": 123,
                    "kes": {
                        "period": 12,
                        "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                    },
                },
                "leaderValue": {},
            },
        }
        Block(Types.praos.value, **kwargs)

    # Invalid height
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "babbage",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "nonce": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "height": -100,
            "slot": 1000,
            "size": {"bytes": 1},
            "transactions": [],  # TODO: Add this type
            "protocol": {
                "version": {"major": 0, "minor": 1, "patch": 2},
            },
            "issuer": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                "vrfVerificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                "operationalCertificate": {
                    "count": 123,
                    "kes": {
                        "period": 12,
                        "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                    },
                },
                "leaderValue": {},
            },
        }
        Block(Types.praos.value, **kwargs)

    # Missing required parameter
    with pytest.raises(ValidationError):
        kwargs = {
            "era": "shelley",
            "id": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "ancestor": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
            "slot": 1000,
            "size": {"bytes": 1},
            "transactions": [],  # TODO: Add this type
            "protocol": {
                "version": {"major": 0, "minor": 1, "patch": 2},
            },
            "issuer": {
                "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                "vrfVerificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                "operationalCertificate": {
                    "count": 123,
                    "kes": {
                        "period": 12,
                        "verificationKey": "45899e8002b27df291e09188bfe3aeb5397ac03546a7d0ead93aa2500860f1af",
                    },
                },
                "leaderValue": {},
            },
        }
        Block(Types.praos.value, **kwargs)


def test_GenesisConfiguration_byron():
    valid_config = {
        "era": "byron",
        "genesis_key_hashes": ["90181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d3"],
        "genesis_delegations": {
            "property1": {
                "issuer": {
                    "verificationKey": "stringstringstringstringstringstringstringstringstringstringstri"
                },
                "delegate": {
                    "verificationKey": "stringstringstringstringstringstringstringstringstringstringstri"
                },
            },
            "property2": {
                "issuer": {
                    "verificationKey": "stringstringstringstringstringstringstringstringstringstringstri"
                },
                "delegate": {
                    "verificationKey": "stringstringstringstringstringstringstringstringstringstringstri"
                },
            },
        },
        "start_time": "2019-08-24T14:15:22Z",
        "initial_funds": {
            "property1": {"ada": {"lovelace": 0}},
            "property2": {"ada": {"lovelace": 0}},
        },
        "initial_vouchers": {
            "property1": {"ada": {"lovelace": 0}},
            "property2": {"ada": {"lovelace": 0}},
        },
        "security_parameter": 18446744073709552000,
        "network_magic": 764824073,
        "updatable_parameters": {
            "scriptVersion": 0,
            "slotDuration": 20000,
            "maxBlockBodySize": {"bytes": 2000000},
            "maxBlockHeaderSize": {"bytes": 2000000},
            "maxTransactionSize": {"bytes": 4096},
            "maxUpdateProposalSize": {"bytes": 700},
            "multiPartyComputationThreshold": "1/50",
            "heavyDelegationThreshold": "3/10000",
            "updateVoteThreshold": "1/1000",
            "updateProposalThreshold": "1/10",
            "updateProposalTimeToLive": 10000,
            "unlockStakeEpoch": 18446744073709551615,
            "softForkInitThreshold": "9/10",
            "softForkMinThreshold": "3/5",
            "softForkDecrementThreshold": "1/20",
            "minFeeConstant": {"ada": {"lovelace": 155381}},
            "minFeeCoefficient": 44,
        },
    }

    genesis_configuration = GenesisConfiguration(**valid_config)
    assert genesis_configuration.era == "byron"
    assert genesis_configuration.genesis_key_hashes == [
        "90181c517a5beadc9c3fe64bf821d3e889a963fc717003ec248757d3"
    ]
    assert genesis_configuration.genesis_delegations.keys() == {"property1", "property2"}
    assert isinstance(genesis_configuration.start_time, datetime)
    assert genesis_configuration.initial_funds.keys() == {"property1", "property2"}
    assert genesis_configuration.initial_vouchers.keys() == {"property1", "property2"}
    assert genesis_configuration.security_parameter == 18446744073709552000
    assert genesis_configuration.network_magic == 764824073
    assert isinstance(genesis_configuration.updatable_parameters, BootstrapProtocolParameters)

    # Optional ProtocolParameters
    valid_config.pop("updatable_parameters")
    genesis_configuration = GenesisConfiguration(**valid_config)
    assert isinstance(genesis_configuration, GenesisConfiguration)

    # Missing required parameter
    invalid_config = valid_config.copy()
    invalid_config.pop("era")
    with pytest.raises(TypeError):
        GenesisConfiguration(**invalid_config)


def test_GenesisConfiguration_shelley():
    valid_config = {
        "era": "shelley",
        "start_time": "1864-05-14T12:52:58Z",
        "network_magic": 5,
        "network": "mainnet",
        "active_slots_coefficient": "1338933938837/2500000000000",
        "security_parameter": 6,
        "epoch_length": 10,
        "slots_per_kes_period": 8,
        "max_kes_evolutions": 12,
        "slot_length": {"milliseconds": 5500000000},
        "update_quorum": 3,
        "max_lovelace_supply": 6,
        "initial_parameters": {
            "minFeeCoefficient": 334307,
            "minFeeConstant": {"ada": {"lovelace": 32519}},
            "maxBlockBodySize": {"bytes": 2},
            "maxBlockHeaderSize": {"bytes": 4},
            "maxTransactionSize": {"bytes": 0},
            "stakeCredentialDeposit": {"ada": {"lovelace": 59105}},
            "stakePoolDeposit": {"ada": {"lovelace": 656520}},
            "stakePoolRetirementEpochBound": 9,
            "desiredNumberOfStakePools": 2,
            "stakePoolPledgeInfluence": "6944763091873046659/10000000000000000",
            "minStakePoolCost": {"ada": {"lovelace": 799263}},
            "monetaryExpansion": "2337923867682421/2500000000000000",
            "treasuryExpansion": "494551394454573889/1000000000000000000",
            "federatedBlockProductionRatio": "1/10",
            "extraEntropy": "7cf8ec2696b93e93d87cb1d59cbe9343aefc774d68313a1d8a9fc38c5d86009d",
            "minUtxoDepositConstant": {"ada": {"lovelace": 459786}},
            "minUtxoDepositCoefficient": 0,
            "version": {"major": 9, "minor": 5},
        },
        "initial_delegates": [
            {
                "issuer": {"id": "071ca6ccedf196ddca60d8522e04f670a715a5a0da17745ed7af8028"},
                "delegate": {
                    "id": "888e976949b0b94df754d2776bc0e95347d702c36bef13dabb7cb344",
                    "vrfVerificationKeyHash": "e66a75f9333b7ac5039b213b03c10e72b50f240236f04ae1660afa8e2dbf3ee6",
                },
            },
            {
                "issuer": {"id": "50d964d38f17b9a90884fb467798b6af0aab2b3a26f6a4b64e97c06e"},
                "delegate": {
                    "id": "2f424be0813e119d4b8b17138249ca9472ae7d7e3691a23a9219a18f",
                    "vrfVerificationKeyHash": "02bc1caadd2c65e15d20e942bd60d18035536f2a6c69aaaa8f15292a70e183b4",
                },
            },
            {
                "issuer": {"id": "a75b569648ca237f9f8627938492b19540a9b3501a1f8de3d95b42df"},
                "delegate": {
                    "id": "9d094b662fd67b342363cdc5f325717a002c4e9f2981b23e4a3f1bed",
                    "vrfVerificationKeyHash": "5245b1fd132c574ace1b7745386b7862b4bb9c7a9dddeb99f825e54d3d8840de",
                },
            },
        ],
        "initial_funds": {
            "addr1z9wpmfm7tvvw8rzclyj4r03vfemjcm2gnnphxny2524qz22gtfcw7qeul7s0rhuyatk2dmhjh43f20y8x2rt68mymqcshvjs4y": {
                "ada": {"lovelace": 841057}
            },
            "addr_test1zzuht8ecma93glf73engw92g7ckyve2tex49rvl5lzm0uzggx4mxyh5ud6yq5pzfcg5pl7gs6msyq6a87j0lkwuhs67qxlwfkc": {
                "ada": {"lovelace": 834989}
            },
            "addr1v8nwyquhku0anazevzmvczuh7x5kwzaqxaq60gz2mkq0zkg34ce0f": {
                "ada": {"lovelace": 535223}
            },
            "addr_test1qr4rjx0aeah96mfx3el5ay3a4cp87ckh9xpr5wcmhd2q08xprvnuhndwmzq70gw25azaxwya8j7mltthcr9xglk3kgcssydgnf": {
                "ada": {"lovelace": 750137}
            },
            "addr_test1wrh49wsfwua290nx025w36yqn5mrrgxpzlxp39per6rl58s0suyje": {
                "ada": {"lovelace": 35407}
            },
        },
        "initial_stake_pools": {
            "stakePools": {
                "pool1426r0jwht0kv596l4r7pvhaqcj4c7afu8u0tc8r5mawdj5wkcwm": {
                    "id": "pool1426r0jwht0kv596l4r7pvhaqcj4c7afu8u0tc8r5mawdj5wkcwm",
                    "vrfVerificationKeyHash": "32289a42d0f9a33f0a1a0a0db7e310e07eb9073c1a3386af633a18ac164ed4de",
                    "pledge": {"ada": {"lovelace": 164537}},
                    "cost": {"ada": {"lovelace": 565775}},
                    "margin": "184631597529/200000000000",
                    "rewardAccount": "stake_test17z4f3g8kzeywj4wece4xtszj3wgnl73d4cx63u0svm0v38clcpjyn",
                    "owners": [
                        "24c5583cba1ebdc321de0a088812de63d8f63dae6aa377706e0d7fc5",
                        "71ccf716b9fd239cbff71c336fac46c04947d542bd216a8c3fd26412",
                        "a7022530cb18c3723fabbe8e9c42bc82d36c33e3355eeb993bb89469",
                    ],
                    "relays": [
                        {
                            "type": "hostname",
                            "hostname": "RwbiBR7jFIukrTsv7PrUW05gJydEfSlC3.com",
                        },
                        {"type": "ipAddress", "ipv4": "0.0.0.6", "ipv6": "0:8:0:1::3"},
                    ],
                    "metadata": {
                        "url": "https://PBAzsHajgW0rK1JV5eTCLCpecLrLxi8AG-z4.com",
                        "hash": "36",
                    },
                },
                "pool1f224p497547tz5xdy77rxmz5dc5ntr93xqm3dym8s9xqc9htk20": {
                    "id": "pool1f224p497547tz5xdy77rxmz5dc5ntr93xqm3dym8s9xqc9htk20",
                    "vrfVerificationKeyHash": "a382f459aa026bc31762cdf3019e260a6c1e849dc0fcc17475a8f9bbe21b99ce",
                    "pledge": {"ada": {"lovelace": 334945}},
                    "cost": {"ada": {"lovelace": 56740}},
                    "margin": "65983/200000",
                    "rewardAccount": "stake_test17p6wetcm4treyrczfl34qknmhuc4fke82qj2evmckvfne9qqem4cc",
                    "owners": [
                        "2769036cf196b9a5c81acf0203d86bdfc6197d93fbd29fc8998e2920",
                        "2aa79a3d455ca1d25b832d32b3e77c4f3cec0ffb3c86076ec2f4e040",
                        "3e17b471d1c0d937bc6ffded4a89f72cd5da7771de12fcbf8d44d3db",
                        "aa6d101aacf21268f05c781be0883fd0d4455613547d3b0c00701d37",
                    ],
                    "relays": [
                        {"type": "hostname", "hostname": "73aNjh.com", "port": 5},
                        {"type": "ipAddress", "ipv4": "0.0.0.5"},
                        {
                            "type": "hostname",
                            "hostname": "LcgxHyKpFChdVoH74KKsYCrZwIh3N57.com",
                            "port": 6,
                        },
                        {
                            "type": "hostname",
                            "hostname": "JbKSORwhO2oQAuE5mN8V40tgOOj15wXheruygSDgvIoxmhJrG.com",
                            "port": 7,
                        },
                        {"type": "hostname", "hostname": "hgDgYDlteF1fDzxWwQuIZeB6xWE.com"},
                    ],
                    "metadata": {"url": "https://hw1dJ.com", "hash": "67"},
                },
                "pool1s4cad3jtkyml5e9qwv7qf8jhedhc5wrck6qwak040dtaw0rysyn": {
                    "id": "pool1s4cad3jtkyml5e9qwv7qf8jhedhc5wrck6qwak040dtaw0rysyn",
                    "vrfVerificationKeyHash": "09e7a2aff4f9d88bff2d8282901138cc11ff63c7c35956916236e369c04ca53c",
                    "pledge": {"ada": {"lovelace": 228853}},
                    "cost": {"ada": {"lovelace": 3368}},
                    "margin": "53/100",
                    "rewardAccount": "stake17xa2ac634h9dvg2l0f8qcagmk5j0ktfh06hvslv4wxlsumg5nr854",
                    "owners": [
                        "2c77808a0001a0fcdb3ed6ce92244c54840fa4f7c9de47371838ddfa",
                        "4af4dfc5a82720536fc190d4d128db954bcddbab7f30d7167d916d02",
                        "938f183de817cf3fe4fd2c3f1937b738aa4c9c5b9b5206d65fafd85d",
                    ],
                    "relays": [
                        {"type": "hostname", "hostname": "y0pcPeUIh.com", "port": 5},
                        {
                            "type": "hostname",
                            "hostname": "SxqqQlUfHfN0jc6BXgO-T3Z9k.com",
                            "port": 1,
                        },
                        {"type": "ipAddress", "ipv6": "0:1:0:6:0:8:0:8", "port": 4},
                    ],
                    "metadata": {"url": "https://3LF4CEoIywCrowxJg.com", "hash": "8ef7"},
                },
                "pool19r2pczfq2dsphsgdkpaf8a297ejhvve654l4e37dmawmslkyncp": {
                    "id": "pool19r2pczfq2dsphsgdkpaf8a297ejhvve654l4e37dmawmslkyncp",
                    "vrfVerificationKeyHash": "ad83c09143c7557bdf2347fc1347909ee18baa7a8af74c3564777c2c3b6156f4",
                    "pledge": {"ada": {"lovelace": 440636}},
                    "cost": {"ada": {"lovelace": 778485}},
                    "margin": "1/1",
                    "rewardAccount": "stake_test17q99v6j2rd94cs4ttpc9yrm94rllc3ncu25tkxxrdw2upgq06qauy",
                    "owners": [
                        "0abf0d596e8c41c982bd786d678e3b682ecc19ec99ee9a6db32ae4ad",
                        "33e56066c25ee952b1819e4a4e1479c4b76dfe5a7832c0f3d93e05fb",
                        "b47d4d14662f8bdb2a3e24bdb050dccb29281c8e955ff89427ad0f0f",
                    ],
                    "relays": [
                        {
                            "type": "hostname",
                            "hostname": "VxRy-dXxjkx5h3v3b6rs4tKwdVIzkADgLnqx4a90R.com",
                            "port": 8,
                        },
                        {
                            "type": "hostname",
                            "hostname": "dLS1PsH5ugzRm50zC0Nh8lcW7D1MFUXQwptSoNl5.com",
                        },
                    ],
                    "metadata": {"url": "https://10t.q.r9H2.com", "hash": "ead416569a"},
                },
                "pool1kqmu4u6zdjfnq7kexlutyzatsnsappe0z4hlzynynam7ut93e9u": {
                    "id": "pool1kqmu4u6zdjfnq7kexlutyzatsnsappe0z4hlzynynam7ut93e9u",
                    "vrfVerificationKeyHash": "5caa1ab0c9fc65d8bb55abad44d1c9abc7e544ec46934c5714364899741a7f66",
                    "pledge": {"ada": {"lovelace": 840868}},
                    "cost": {"ada": {"lovelace": 861373}},
                    "margin": "899/1000",
                    "rewardAccount": "stake178mq5uq0pdwsqxgxqrkkehffupun584warusx4p3kk9t7scn7a9zm",
                    "owners": [
                        "83203768841eefd9a04408180cfc2ff4a0cd4103bf161f71062e4395",
                        "bbb49b1cd27e36f051a35bb1de62d2bd5987e99123abf96a6e386247",
                        "d5029d809f698e41168e378a4304be24a5dca539b40d8a46d3d391dd",
                    ],
                    "relays": [
                        {"type": "hostname", "hostname": "GDLiIPxGMSudxNn.com", "port": 2},
                        {
                            "type": "hostname",
                            "hostname": "rw.yqNDbCKLNuD5KOU165XZgbbu.com",
                            "port": 5,
                        },
                        {"type": "ipAddress", "ipv6": "0:0:0:2::"},
                        {"type": "hostname", "hostname": "BpggE8Aq8hQRoelEM1Mc5eB.com"},
                        {
                            "type": "hostname",
                            "hostname": "ltBJLzo0oFxAbUsfsBbU9mK31e2JBdK.XGNL.40I2JJ8Zxu9RB8.com",
                        },
                        {
                            "type": "ipAddress",
                            "ipv4": "0.0.0.1",
                            "ipv6": "0:4:0:7:0:4:0:1",
                            "port": 7,
                        },
                    ],
                    "metadata": {
                        "url": "https://gSkhfClLO4qMnlKUh9Ykk26.p4OI84uvRZ12TVqCWX4.com",
                        "hash": "7326",
                    },
                },
                "pool1evy8aa4dpmchen4yspjsc9x6j38jtxjqz556getkj860jnpqa4w": {
                    "id": "pool1evy8aa4dpmchen4yspjsc9x6j38jtxjqz556getkj860jnpqa4w",
                    "vrfVerificationKeyHash": "f9f3f88ee69184d618951f8226b4ca6c8df7ece09a53f5233af95df731891735",
                    "pledge": {"ada": {"lovelace": 729846}},
                    "cost": {"ada": {"lovelace": 559218}},
                    "margin": "1706179828357/2500000000000",
                    "rewardAccount": "stake_test1urqm8w75x8pkgw0kwytxullk9ccrk20n9ytwdvl33afdkwc5t59j4",
                    "owners": [
                        "097303a5ae59a8c95923a021d6ea259f16e70769fc06ad98208750dd",
                        "c405c0a0d6113294d82f539c8ad6ec603e92bd6e2c1b138f4b0866d6",
                    ],
                    "relays": [],
                    "metadata": {
                        "url": "https://uXdMiE3GKmjrbNMkFNQPNzOZjGUZ9WYU.com",
                        "hash": "f5eb",
                    },
                },
            },
            "delegators": {
                "cc4cae00c233b43d6c8319ebdfdcab69e408f9cc2c4f1cf45a889860": "pool1pxn5uhu8kz5qecpgdzfxs68fh58z7q0ld56qsn3us94asvtrw2x",
                "a9929526d7eeb5d47d5853ffa17dfd013c2387c31c3e3d8732856c2f": "pool1322heh7ry3u5h53ahzx443t8n6tymv5h933twrq3z7cyymvygsc",
                "7bf741aa5dfc567dc3ba556e76ec36d36ccaa5b81343c7277f704050": "pool19kzx76r6a0p5vgc5jer4amy7h4jgqpru6ulfgjm94ke2yleswaf",
                "f3a63578372c8e780c05dd48eeae8b6c8b7f111bddf05c729c0373e2": "pool1sj3xltrqh6yccm99sllucdl66pd4udq7wxqdfx35y2ld5jj3lqv",
                "bf5267ca1a97d0629c08493d0ad641ba4ec0586a77a2074e907c34a9": "pool1wjkugue635ckgyfwjvur8l9uyqvtv4y0qdql4vzw8uxa6z3rhdz",
            },
        },
    }

    genesis_configuration = GenesisConfiguration(**valid_config)
    assert genesis_configuration.era == "shelley"
    assert isinstance(genesis_configuration.initial_parameters, ProtocolParameters)

    # Missing required parameter
    invalid_config = valid_config.copy()
    invalid_config.pop("era")
    with pytest.raises(TypeError):
        GenesisConfiguration(**invalid_config)


def test_script():
    native_script = Script(
        language="native",
        cbor="3c07030e36bfff7cd2f004356ef320f3fe3c07030e7cd2f004356437",
        json={
            "clause": "all",
            "from": [
                {
                    "clause": "signature",
                    "from": "ec09e5293d384637cd2f004356ef320f3fe3c07030e36bfffe67e2e2",
                }
            ],
        },
    )
    assert native_script.language == "native"

    plutus_v2_script = Script(
        language="plutus:v2",
        cbor="3c07030e36bfff7cd2f004356ef320f3fe3c07030e7cd2f004356437",
    )
    assert plutus_v2_script.language == "plutus:v2"



from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryEraStart(client):
    id_str = "My ID string"
    time, slot, epoch, id = client.query_era_start.execute(id_str)

    assert isinstance(time, int)
    assert isinstance(slot, int)
    assert isinstance(epoch, int)
    assert id == id_str



from pydantic.v1 import ValidationError
import time

from ogmios.client import Client
from ogmios.datatypes import Era

"""
This module contains helpful utilities for working with Ogmios.
"""


class GenesisParameters:
    """A class representing the genesis parameters of the blockchain. Each era has its own genesis
    configuration, whose parameters are additive to all previous eras. Therefore, to get the full set
    of genesis parameters, we need to query all eras up to the present and combine their parameters.

    :param latest_era: The latest era of the blockchain for which to compile genesis parameters
    :type latest_era: Era
    """

    def __init__(
        self,
        client: Client,
        latest_era: Era = Era.conway,
    ):
        # Query the genesis parameters for each era up to the latest era
        for i in range(len(Era)):
            era = Era.by_index(i)
            if Era.is_genesis_era(era):
                self.era = era.value
                genesis_parameters, _ = client.query_genesis_configuration.execute(era.value)

                # Unpack the genesis parameters into the class
                for key, value in genesis_parameters.__dict__.items():
                    setattr(self, key, value)

            if Era.by_index(i) == latest_era:
                break


def get_current_era(client: Client) -> Era:
    """
    Get the current era of the blockchain

    :param client: The Ogmios client object
    :type client: Client
    :return: The current era of the blockchain
    :rtype: Era
    """
    era_summaries, _ = client.query_era_summaries.execute()
    return Era.by_index(len(era_summaries) - 1)


def wait_for_empty_mempool(client: Client, timeout_s=60) -> None:
    """
    Wait for the mempool to be empty

    :param client: The Ogmios client object
    :type client: Client
    """
    start_time = time.time()
    while True:
        client.acquire_mempool.execute()
        tx, id = client.next_transaction.execute()
        client.release_mempool.execute()
        if tx is None:
            break
        if time.time() - start_time > timeout_s:
            raise TimeoutError("Mempool did not empty within the timeout period")
        time.sleep(1)


def get_mempool_transactions(client: Client) -> list:
    """
    Get the contents of the mempool

    :param client: The Ogmios client object
    :type client: Client
    :return: The contents of the mempool (list of transactions)
    :rtype: list
    """
    client.acquire_mempool.execute()
    mempool_txs = []
    while True:
        tx, id = client.next_transaction.execute()
        if tx is None:
            break
        mempool_txs.append(tx)
    client.release_mempool.execute()
    return mempool_txs



import pycardano as pyc

network = pyc.Network.TESTNET

# Read keys to memory
psk = pyc.PaymentExtendedSigningKey.load("./tests/test_wallet/test_addr0.skey")
ssk = pyc.StakeExtendedSigningKey.load("./tests/test_wallet/test_stake.skey")

pvk = pyc.PaymentVerificationKey.from_signing_key(psk)
svk = pyc.StakeVerificationKey.from_signing_key(ssk)

# Derive an address from payment verification key and stake verification key
address = pyc.Address(pvk.hash(), svk.hash(), network)

# Create an Ogmios chain context. In this example, we will use preprod network.
context = pyc.OgmiosV6ChainContext("localhost", 1337)

genesis = context.genesis_param

# # Create a transaction builder
# builder = pyc.TransactionBuilder(context)

# # Tell the builder that transaction input will come from a specific address, assuming that there are some ADA and native
# # assets sitting at this address. "add_input_address" could be called multiple times with different address.
# builder.add_input_address(address)

# # Get all UTxOs currently sitting at this address
# utxos = context.utxos(address)

# # We can also tell the builder to include a specific UTxO in the transaction.
# # Similarly, "add_input" could be called multiple times.
# builder.add_input(utxos[0])

# builder.add_output(
#     pyc.TransactionOutput(
#         pyc.Address.from_primitive(
#             "addr_test1vrm9x2zsux7va6w892g38tvchnzahvcd9tykqf3ygnmwtaqyfg52x"
#         ),
#         pyc.Value.from_primitive(
#             [
#                 1500000,
#             ]
#         ),
#     )
# )

# # We can add multiple outputs, similar to what we can do with inputs.
# # Send 2 ADA and a native asset (CHOC) in quantity of 200 to ourselves
# builder.add_output(
#     pyc.TransactionOutput(
#         address,
#         pyc.Value.from_primitive(
#             [
#                 2000000,
#             ]
#         ),
#     )
# )

# # Create final signed transaction
# signed_tx = builder.build_and_sign([psk], change_address=address)
# print(signed_tx)

# # Submit signed transaction to the network
# context.submit_tx(signed_tx)



from tests.ogmios.test_fixtures import client


def test_QueryLedgerTip(client):
    id_str = "My ID string"
    result, id = client.query_rewards_provenance.execute(id_str)

    assert isinstance(result, dict)
    expected_keys = [
        'desiredNumberOfStakePools',
        'stakePoolPledgeInfluence',
        'totalRewardsInEpoch',
        'activeStakeInEpoch',
        'stakePools',
    ]
    for key in expected_keys:
        assert key in result.keys()
    assert id == id_str



import pytest

from tests.ogmios.test_fixtures import client, era
from ogmios.datatypes import Era


def test_QueryConstitutionalCommittee(client, era):
    if Era.get_index(era) < Era.get_index(Era.conway):
        pytest.skip("Test only valid for conway era")

    id_str = "My ID string"
    result, id = client.query_constitutional_committee.execute(id_str)

    if result:
        assert isinstance(result, dict)
        assert result.keys() == {"members", "quorum"}
    assert id == id_str



from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryStakePools(client):
    id_str = "My ID string"
    stake_dist, _ = client.query_live_stake_distribution.execute()

    test_pools = list(stake_dist.keys())[:2]
    pool_summaries, id = client.query_stake_pools.execute(test_pools, id_str)

    assert isinstance(pool_summaries, dict)
    assert id == id_str

    summary = pool_summaries[test_pools[0]]
    for key in summary.keys():
        assert key in [
            'id',
            'vrfVerificationKeyHash',
            'pledge',
            'cost',
            'margin',
            'rewardAccount',
            'owners',
            'relays',
            'metadata',
        ]



from datetime import datetime

from ogmios.datatypes import Ada
from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryProjectedRewards_key(client):
    # This test relies on this address having a balance on the testnet
    stake_key = "stake_test1urlvxquq2pm8teqed7ek4sgjuc0zrx07z9czmy5fvwey92cq8c5dh"
    id_str = "My ID string"
    projection, id = client.query_projected_rewards.execute(
        stake=[Ada(10000)], keys=[stake_key], id=id_str
    )

    assert isinstance(projection, dict)
    assert id == id_str


def test_QueryProjectedRewards_script(client):
    # This test relies on this address having a balance on the testnet
    script_key = "stake_test1urlvxquq2pm8teqed7ek4sgjuc0zrx07z9czmy5fvwey92cq8c5dh"
    id_str = "My ID string"
    projection, id = client.query_projected_rewards.execute(
        stake=[Ada(10000)], scripts=[script_key], id=id_str
    )

    assert isinstance(projection, dict)
    assert id == id_str



import pytest
import json

from ogmios.errors import InvalidMethodError, InvalidResponseError, ResponseError
from ogmios.datatypes import Origin, Point, GenesisConfiguration, Era

from tests.ogmios.test_fixtures import client, era

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryGenesisConfiguration_byron(client):
    id_str = "My ID string"
    genesis_configuration, id = client.query_genesis_configuration.execute(Era.byron.value, id_str)

    assert isinstance(genesis_configuration, GenesisConfiguration)
    assert id == id_str


def test_QueryGenesisConfiguration_shelley(client, era):
    if Era.get_index(era) < Era.get_index(Era.shelley):
        pytest.skip("Test only valid for shelley era")

    id_str = "My ID string"
    genesis_configuration, id = client.query_genesis_configuration.execute(
        Era.shelley.value, id_str
    )

    assert isinstance(genesis_configuration, GenesisConfiguration)
    assert id == id_str


def test_QueryGenesisConfiguration_alonzo(client, era):
    if Era.get_index(era) < Era.get_index(Era.alonzo):
        pytest.skip("Test only valid for alonzo era")

    id_str = "My ID string"
    genesis_configuration, id = client.query_genesis_configuration.execute(Era.alonzo.value, id_str)

    assert isinstance(genesis_configuration, GenesisConfiguration)
    assert id == id_str


def test_QueryGenesisConfiguration_conway(client, era):
    if Era.get_index(era) < Era.get_index(Era.conway):
        pytest.skip("Test only valid for conway era")

    id_str = "My ID string"
    genesis_configuration, id = client.query_genesis_configuration.execute(Era.conway.value, id_str)

    assert isinstance(genesis_configuration, GenesisConfiguration)
    assert id == id_str



import pytest

from ogmios.errors import ResponseError
from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_SizeOfMempool(client):
    id_str = "My ID string"
    with pytest.raises(ResponseError):
        # Try getting mempool size before acquiring a snapshot
        client.size_of_mempool.execute()

    # Acquire a mempool snapshot
    client.acquire_mempool.execute()
    max_capacity, current_size, num_transactions, id = client.size_of_mempool.execute(id_str)
    assert isinstance(max_capacity, int)
    assert isinstance(current_size, int)
    assert isinstance(num_transactions, int)
    assert max_capacity > 0
    assert id == id_str
    if current_size > 0:
        assert num_transactions > 0
    else:
        assert num_transactions == 0



from tests.ogmios.test_fixtures import client
from ogmios.datatypes import Ada


def test_QueryTreasuryAndReserves(client):
    id_str = "My ID string"
    treasury, reserves, id = client.query_treasury_and_reserves.execute(id_str)
    assert isinstance(treasury, Ada)
    assert isinstance(reserves, Ada)
    assert id == id_str



from __future__ import annotations

from typing import Any, Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ogmios.client import Client

from ogmios.errors import InvalidMethodError, InvalidResponseError, ResponseError
from ogmios.logger import logger
import ogmios.model.ogmios_model as om
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


class QueryStakePools:
    """Ogmios method to list of all stake pool identifiers currently registered and active.

    NOTE: This class is not intended to be used directly. Instead, use the Client.query_stake_pools
    method.

    :param client: The client to use for the request.
    :type client: Client
    """

    def __init__(self, client: Client):
        self.client = client
        self.method = mm.Method.queryLedgerState_stakePools.value

    def execute(self, stake_pools: list[str], id: Optional[Any] = None) -> (dict, Optional[Any]):
        """Send and receive the request.

        :param stake_pools: The list of stake pool bech32 IDs to query.
        :type points: list[str]
        :param id: The ID of the request.
        :type id: Any
        :return: Dict of stake pool summaries and ID of the response.
        :rtype: (dict, Optional[Any])
        """
        self.send(stake_pools, id)
        return self.receive()

    def send(self, stake_pools: list[str], id: Optional[Any] = None) -> None:
        """Send the request.

        :param stake_pools: The list of stake pool bech32 IDs to query.
        :type points: list[str]
        :param id: The ID of the request.
        :type id: Any
        """
        params = om.Params6(stakePools=[om.StakePool(id=stake_pool) for stake_pool in stake_pools])
        pld = om.QueryLedgerStateStakePools(
            jsonrpc=self.client.rpc_version,
            method=self.method,
            params=params,
            id=id,
        )
        self.client.send(pld.json())

    def receive(self) -> (dict, Optional[Any]):
        """Receive the response.

        :return: Dict of stake pool summaries and ID of the response.
        :rtype: (dict, Optional[Any])
        """
        response = self.client.receive()
        return self._parse_QueryStakePools_response(response)

    @staticmethod
    def _parse_QueryStakePools_response(
        response: dict,
    ) -> (dict, Optional[Any]):
        if response.get("method") != mm.Method.queryLedgerState_stakePools.value:
            raise InvalidMethodError(f"Incorrect method for query_stake_pool response: {response}")

        if response.get("error"):
            raise ResponseError(f"Ogmios responded with error: {response}")

        if result := response.get("result"):
            id: Optional[Any] = response.get("id")
            logger.info(
                f"""Parsed query_stake_pool response:
        Stake Pool Summaries = {result}
        ID = {id}"""
            )
            return result, id

        raise InvalidResponseError(f"Failed to parse query_stake_pool response: {response}")



import time
import ogmios

"""
This example shows how to use the ogmios-python library to view new blocks as they are added to the
blockchain.
"""


def live_block_viewer():
    with ogmios.Client() as client:
        # Set chain pointer to origin
        point, tip, id = client.find_intersection.execute([ogmios.Origin()])

        # Now set chain pointer to tip
        _, _, _ = client.find_intersection.execute([tip.to_point()])

        # Tail blockchain as new blocks come in beyond the current tip
        while True:
            direction, tip, point, _ = client.next_block.execute()
            if direction == ogmios.Direction.forward:
                print(f"New block: {point}")
            time.sleep(1)


if __name__ == "__main__":
    live_block_viewer()



import time
import pytest
import pycardano as pyc

from ogmios.errors import ResponseError
from tests.ogmios.test_fixtures import (
    client,
    chain_context,
    test_psk,
    test_ssk,
    test_pvk,
    test_svk,
    test_address,
)

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


@pytest.mark.skip
def test_SubmitTransaction(client, chain_context, test_psk, test_pvk, test_svk, test_address):
    # Make sure mempool is empty
    wait_for_mempool_empty(client)

    tx_builder = pyc.TransactionBuilder(chain_context)
    tx_builder.add_input_address(test_address)

    utxos = chain_context.utxos(test_address)
    utxos = chain_context.utxos(test_address)
    tx_utxo = None
    for utxo in utxos:
        if utxo.output.amount.coin > 1_000_000:
            tx_utxo = utxo
            break

    if tx_utxo is None:
        pytest.fail("No suitable UTxO found")

    tx_builder.add_input(tx_utxo)
    tx_builder.add_output(
        pyc.TransactionOutput(
            test_address,
            pyc.Value.from_primitive([1_000_000]),
        )
    )
    signed_tx = tx_builder.build_and_sign([test_psk], change_address=test_address)

    resp = client.submit_transaction.execute(signed_tx.to_cbor_hex())
    assert resp[0] == str(signed_tx.id)
    wait_for_mempool_empty(client)
    assert True


def wait_for_mempool_empty(client) -> None:
    # Wait for the transaction to be removed from the mempool
    while True:
        client.acquire_mempool.execute()
        tx, id = client.next_transaction.execute()
        client.release_mempool.execute()
        if tx is None:
            break
        time.sleep(1)



from ogmios.datatypes import Origin, Point

from tests.ogmios.test_fixtures import client

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_QueryNetworkTip(client):
    # Get tip
    _, tip, _ = client.find_intersection.execute([Origin()])

    id_str = "My ID string"
    queried_tip, id = client.query_network_tip.execute(id_str)

    assert isinstance(queried_tip, Point)
    assert queried_tip == tip.to_point()
    assert id == id_str



import pytest
from datetime import datetime
from pydantic.v1.error_wrappers import ValidationError

from tests.ogmios.test_fixtures import client

from ogmios.utils import GenesisParameters, get_current_era, wait_for_empty_mempool
from ogmios.datatypes import Era

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


def test_GenesisParameters(client):
    for i in range(len(Era)):
        genesis = GenesisParameters(client, Era.by_index(i))
        assert hasattr(genesis, "genesis_delegations")
        assert hasattr(genesis, "genesis_key_hashes")
        assert hasattr(genesis, "initial_funds")
        assert hasattr(genesis, "initial_vouchers")
        assert hasattr(genesis, "start_time")
        assert hasattr(genesis, "updatable_parameters")
        if genesis.era == "byron":
            continue

        assert hasattr(genesis, "active_slots_coefficient")
        assert hasattr(genesis, "epoch_length")
        assert hasattr(genesis, "initial_delegates")
        assert hasattr(genesis, "initial_parameters")
        assert hasattr(genesis, "initial_stake_pools")
        assert hasattr(genesis, "max_kes_evolutions")
        assert hasattr(genesis, "max_lovelace_supply")
        assert hasattr(genesis, "security_parameter")
        assert hasattr(genesis, "slot_length")
        assert hasattr(genesis, "slots_per_kes_period")
        assert hasattr(genesis, "start_time")
        if genesis.era == "shelley":
            continue

        assert hasattr(genesis, "update_quorum")
        if genesis.era == "alonzo":
            continue

        assert hasattr(genesis, "constitution")
        assert hasattr(genesis, "constitutional_committee")
        if genesis.era == "conway":
            continue



from __future__ import annotations

from typing import Any, Optional, Tuple, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ogmios.client import Client

from ogmios.errors import InvalidMethodError, InvalidResponseError
from ogmios.logger import logger
from ogmios.datatypes import Origin, Point, Tip, Block, Direction
import ogmios.response_handler as rh
import ogmios.model.ogmios_model as om
import ogmios.model.model_map as mm

# pyright can't properly parse models, so we need to ignore its type checking
#  (pydantic will still throw errors if we misuse a data type)
# pyright: reportGeneralTypeIssues=false


class NextBlock:
    """Ogmios method to request the next block in the blockchain.

    :param client: The client to use for the request.
    :type client: Client
    """

    def __init__(self, client: Client):
        self.client = client
        self.method = mm.Method.nextBlock.value

    def execute(
        self, id: Optional[Any] = None
    ) -> Tuple[Direction, Tip, Union[Point, Origin, Block], Optional[Any]]:
        """Send and receive the request.

        :param id: The ID of the request.
        :type id: Any
        :return: The direction, tip, point or block or origin, and ID of the response.
        :rtype: (Direction, Tip, Point | Origin | Block, Optional[Any])
        """
        self.send(id)
        return self.receive()

    def send(self, id: Optional[Any] = None) -> None:
        """Send the request.

        :param jsonrpc: The JSON-RPC version to use.
        :type jsonrpc: Jsonrpc
        :param method: The method to use.
        :type method: Method
        :param id: The ID of the request.
        :type id: Any
        """
        pld = om.NextBlock(
            jsonrpc=self.client.rpc_version,
            method=self.method,
            id=id,
        )
        self.client.send(pld.json())

    def receive(self) -> Tuple[Direction, Tip, Union[Point, Origin, Block], Optional[Any]]:
        """Receive a previously requested response.

        :return: The direction, tip, point or block or origin, and ID of the response.
        :rtype: (Direction, Tip, Point | Origin | Block, Optional[Any])
        """
        response = self.client.receive()
        return self._parse_NextBlock_response(response)

    @staticmethod
    def _parse_NextBlock_response(
        response: dict,
    ) -> Tuple[Direction, Tip, Union[Point, Origin, Block], Optional[Any]]:
        if response.get("method") != mm.Method.nextBlock.value:
            raise InvalidMethodError(f"Incorrect method for next_block response: {response}")

        # Successful response will contain direction, tip, and either a block or point
        if result := response.get("result"):
            if result.get("direction") and result.get("tip"):
                direction: Direction = NextBlock._parse_direction(result.get("direction"))
                tip: Union[Tip, Origin] = rh.parse_TipOrOrigin(result.get("tip"))
                id: Optional[Any] = response.get("id")
                if (block_resp := result.get("block")) is not None:
                    block: Block = rh.parse_Block(block_resp)
                    logger.info(
                        f"""Parsed NextBlock response:
        Direction = {direction}
        Tip = {tip}
        Block = {block}
        ID = {id}"""
                    )
                    return direction, tip, block, id
                elif (point_resp := result.get("point")) is not None:
                    point: Union[Point, Origin] = rh.parse_PointOrOrigin(point_resp)
                    logger.info(
                        f"""Parsed NextBlock response:
        Direction = {direction}
        Tip = {tip}
        Point = {point}
        ID = {id}"""
                    )
                    return direction, tip, point, id
        raise InvalidResponseError(f"Failed to parse next_block response: {response}")

    @staticmethod
    def _parse_direction(value: str) -> Direction:
        if value == Direction.forward.value:
            return Direction.forward
        elif value == Direction.backward.value:
            return Direction.backward
        raise InvalidResponseError(
            f"next_block response contains invalid direction parameter: {value}"
        )
